import pytest

from shelloracle.providers.localai import LocalAI


class TestOpenAI:
    @pytest.fixture
    def localai_config(self, set_config):
        config = {
            "shelloracle": {"provider": "LocalAI"},
            "provider": {"LocalAI": {"host": "localhost", "port": 8080, "model": "mistral-openorca"}},
        }
        set_config(config)

    @pytest.fixture
    def localai_instance(self, localai_config):
        return LocalAI()

    def test_name(self):
        assert LocalAI.name == "LocalAI"

    def test_model(self, localai_instance):
        assert localai_instance.model == "mistral-openorca"

    @pytest.mark.asyncio
    async def test_generate(self, mock_asyncopenai, localai_instance):
        result = ""
        async for response in localai_instance.generate(""):
            result += response
        assert result == "head -c 100 /dev/urandom | hexdump -C"



import pytest
import tomlkit

from shelloracle.config import Configuration


@pytest.fixture(autouse=True)
def tmp_shelloracle_home(monkeypatch, tmp_path):
    monkeypatch.setattr("shelloracle.settings.Settings.shelloracle_home", tmp_path)
    return tmp_path


@pytest.fixture
def set_config(monkeypatch, tmp_shelloracle_home):
    config_path = tmp_shelloracle_home / "config.toml"

    def _set_config(config: dict) -> None:
        with config_path.open("w") as f:
            tomlkit.dump(config, f)
        configuration = Configuration(config_path)
        monkeypatch.setattr("shelloracle.config._config", configuration)

    yield _set_config

    config_path.unlink()



from collections.abc import AsyncIterator

from openai import APIError, AsyncOpenAI

from shelloracle.providers import Provider, ProviderError, Setting, system_prompt


class OpenAI(Provider):
    name = "OpenAI"

    api_key = Setting(default="")
    model = Setting(default="gpt-3.5-turbo")

    def __init__(self):
        if not self.api_key:
            msg = "No API key provided"
            raise ProviderError(msg)
        self.client = AsyncOpenAI(api_key=self.api_key)

    async def generate(self, prompt: str) -> AsyncIterator[str]:
        try:
            stream = await self.client.chat.completions.create(
                model=self.model,
                messages=[{"role": "system", "content": system_prompt}, {"role": "user", "content": prompt}],
                stream=True,
            )
            async for chunk in stream:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        except APIError as e:
            msg = f"Something went wrong while querying OpenAI: {e}"
            raise ProviderError(msg) from e



from __future__ import annotations

import logging
import sys
from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, Any

from yaspin.spinners import SPINNERS_DATA

from shelloracle.settings import Settings

if TYPE_CHECKING:
    from pathlib import Path

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

logger = logging.getLogger(__name__)


class Configuration(Mapping):
    def __init__(self, filepath: Path) -> None:
        """ShellOracle application configuration

        :param filepath: Path to the configuration file
        :raises FileNotFoundError: if the configuration file does not exist
        """
        self.filepath = filepath
        with filepath.open("rb") as config_file:
            self._config = tomllib.load(config_file)

    def __getitem__(self, key: str) -> Any:
        return self._config[key]

    def __len__(self) -> int:
        return len(self._config)

    def __iter__(self) -> Iterator[Any]:
        return iter(self._config)

    def __str__(self):
        return f"Configuration({self._config})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def provider(self) -> str:
        return self["shelloracle"]["provider"]

    @property
    def spinner_style(self) -> str | None:
        style = self["shelloracle"].get("spinner_style", None)
        if not style:
            return None
        if style not in SPINNERS_DATA:
            logger.warning("invalid spinner style: %s", style)
            return None
        return style


_config: Configuration | None = None


def initialize_config() -> None:
    """Initialize the configuration file

    :raises RuntimeError: if the config is already initialized
    :raises FileNotFoundError: if the config file is not found
    """
    global _config  # noqa: PLW0603
    if _config:
        msg = "Configuration already initialized"
        raise RuntimeError(msg)
    filepath = Settings.shelloracle_home / "config.toml"
    _config = Configuration(filepath)


def get_config() -> Configuration:
    """Returns the global configuration object.

    :return: the global configuration
    :raises RuntimeError: if the configuration is not initialized
    """
    if _config is None:
        msg = "Configuration not initialized"
        raise RuntimeError(msg)
    return _config



from __future__ import annotations

import pytest

from shelloracle.config import get_config, initialize_config


class TestConfiguration:
    @pytest.fixture
    def default_config(self, set_config):
        config = {
            "shelloracle": {"provider": "Ollama", "spinner_style": "earth"},
            "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
        }
        set_config(config)
        return config

    def test_initialize_config(self, default_config):
        with pytest.raises(RuntimeError):
            initialize_config()

    def test_from_file(self, default_config):
        assert get_config() == default_config

    def test_getitem(self, default_config):
        for key in default_config:
            assert default_config[key] == get_config()[key]

    def test_len(self, default_config):
        assert len(default_config) == len(get_config())

    def test_iter(self, default_config):
        assert list(iter(default_config)) == list(iter(get_config()))

    def test_str(self, default_config):
        assert str(get_config()) == f"Configuration({default_config})"

    def test_repr(self, default_config):
        assert repr(default_config) == str(default_config)

    def test_provider(self, default_config):
        assert get_config().provider == "Ollama"

    def test_spinner_style(self, default_config):
        assert get_config().spinner_style == "earth"

    def test_no_spinner_style(self, caplog, set_config):
        config_dict = {
            "shelloracle": {"provider": "Ollama"},
            "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
        }
        set_config(config_dict)
        assert get_config().spinner_style is None
        assert "invalid spinner style" not in caplog.text

    def test_invalid_spinner_style(self, caplog, set_config):
        config_dict = {
            "shelloracle": {"provider": "Ollama", "spinner_style": "invalid"},
            "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
        }
        set_config(config_dict)
        assert get_config().spinner_style is None
        assert "invalid spinner style" in caplog.text



from __future__ import annotations

import os
import sys
from unittest.mock import MagicMock, call

import pytest
from yaspin.spinners import Spinners

from shelloracle.shelloracle import get_query_from_pipe, spinner


@pytest.fixture
def mock_yaspin(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr("shelloracle.shelloracle.yaspin", mock)
    return mock


@pytest.fixture
def mock_config(monkeypatch):
    config = MagicMock()
    monkeypatch.setattr("shelloracle.config._config", config)
    return config


@pytest.mark.parametrize(("spinner_style", "expected"), [(None, call()), ("earth", call(Spinners.earth))])
def test_spinner(spinner_style, expected, mock_config, mock_yaspin):
    mock_config.spinner_style = spinner_style
    spinner()
    assert mock_yaspin.call_args == expected


def test_spinner_fail(mock_yaspin, mock_config):
    mock_config.spinner_style = "not a spinner style"
    with pytest.raises(AttributeError):
        spinner()


@pytest.mark.parametrize(
    ("isatty", "readlines", "expected"), [(True, None, None), (False, [], None), (False, ["what is up"], "what is up")]
)
def test_get_query_from_pipe(isatty, readlines, expected, monkeypatch):
    monkeypatch.setattr(os, "isatty", lambda _: isatty)
    monkeypatch.setattr(sys.stdin, "readlines", lambda: readlines)
    assert get_query_from_pipe() == expected


def test_get_query_from_pipe_fail(monkeypatch):
    monkeypatch.setattr(os, "isatty", lambda _: False)
    monkeypatch.setattr(sys.stdin, "readlines", lambda: ["what is up", "what is down"])
    with pytest.raises(ValueError, match="Multi-line input is not supported"):
        get_query_from_pipe()



from unittest.mock import MagicMock

import pytest


def split_with_delimiter(string, delim):
    result = []
    last_split = 0
    for index, character in enumerate(string):
        if character == delim:
            result.append(string[last_split : index + 1])
            last_split = index + 1
    if last_split != len(string):
        result.append(string[last_split:])
    return result


@pytest.fixture
def mock_asyncopenai(monkeypatch):
    class AsyncChatCompletionIterator:
        def __init__(self, answer: str):
            self.answer_index = 0
            self.answer_deltas = split_with_delimiter(answer, " ")

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.answer_index >= len(self.answer_deltas):
                raise StopAsyncIteration
            answer_chunk = self.answer_deltas[self.answer_index]
            self.answer_index += 1
            choice = MagicMock()
            choice.delta.content = answer_chunk
            chunk = MagicMock()
            chunk.choices = [choice]
            return chunk

    async def mock_acreate(*args, **kwargs):
        return AsyncChatCompletionIterator("head -c 100 /dev/urandom | hexdump -C")

    monkeypatch.setattr("openai.resources.chat.AsyncCompletions.create", mock_acreate)



import argparse
import logging
import sys
from importlib.metadata import version

from shelloracle import shelloracle
from shelloracle.config import initialize_config
from shelloracle.settings import Settings
from shelloracle.tty_log_handler import TtyLogHandler

logger = logging.getLogger(__name__)


def configure_logging():
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    file_formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    file_handler = logging.FileHandler(Settings.shelloracle_home / "shelloracle.log")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(file_formatter)

    tty_formatter = logging.Formatter("%(message)s")
    tty_handler = TtyLogHandler()
    tty_handler.setLevel(logging.WARNING)
    tty_handler.setFormatter(tty_formatter)

    root_logger.addHandler(file_handler)
    root_logger.addHandler(tty_handler)


def configure():
    # nest this import in a function to avoid expensive module loads
    from shelloracle.bootstrap import bootstrap_shelloracle

    bootstrap_shelloracle()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=f"{__package__} {version(__package__)}")

    subparsers = parser.add_subparsers()
    configure_subparser = subparsers.add_parser("configure", help=f"install {__package__} keybindings")
    configure_subparser.set_defaults(action=configure)

    return parser.parse_args()


def main() -> None:
    args = parse_args()
    configure_logging()

    if action := getattr(args, "action", None):
        action()
        sys.exit(0)

    try:
        initialize_config()
    except FileNotFoundError:
        logger.warning("ShellOracle configuration not found. Run `shor configure` to initialize.")
        sys.exit(1)

    shelloracle.cli()


if __name__ == "__main__":
    main()



from collections.abc import AsyncIterator

from openai import APIError, AsyncOpenAI

from shelloracle.providers import Provider, ProviderError, Setting, system_prompt


class LocalAI(Provider):
    name = "LocalAI"

    host = Setting(default="localhost")
    port = Setting(default=8080)
    model = Setting(default="mistral-openorca")

    @property
    def endpoint(self) -> str:
        return f"http://{self.host}:{self.port}"

    def __init__(self):
        # Use a placeholder API key so the client will work
        self.client = AsyncOpenAI(api_key="sk-xxx", base_url=self.endpoint)

    async def generate(self, prompt: str) -> AsyncIterator[str]:
        try:
            stream = await self.client.chat.completions.create(
                model=self.model,
                messages=[{"role": "system", "content": system_prompt}, {"role": "user", "content": prompt}],
                stream=True,
            )
            async for chunk in stream:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        except APIError as e:
            msg = f"Something went wrong while querying LocalAI: {e}"
            raise ProviderError(msg) from e



import pytest
from pytest_httpx import IteratorStream

from shelloracle.providers.ollama import Ollama


class TestOllama:
    @pytest.fixture
    def ollama_config(self, set_config):
        config = {
            "shelloracle": {"provider": "Ollama"},
            "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
        }
        set_config(config)

    @pytest.fixture
    def ollama_instance(self, ollama_config):
        return Ollama()

    def test_name(self):
        assert Ollama.name == "Ollama"

    def test_host(self, ollama_instance):
        assert ollama_instance.host == "localhost"

    def test_port(self, ollama_instance):
        assert ollama_instance.port == 11434

    def test_model(self, ollama_instance):
        assert ollama_instance.model == "dolphin-mistral"

    def test_endpoint(self, ollama_instance):
        assert ollama_instance.endpoint == "http://localhost:11434/api/generate"

    @pytest.mark.asyncio
    async def test_generate(self, ollama_instance, httpx_mock):
        responses = [
            b'{"response": "cat"}\n',
            b'{"response": " test"}\n',
            b'{"response": "."}\n',
            b'{"response": "py"}\n',
            b'{"response": ""}\n',
        ]
        httpx_mock.add_response(stream=IteratorStream(responses))
        result = ""
        async for response in ollama_instance.generate(""):
            result += response
        assert result == "cat test.py"



import pytest

from shelloracle.providers.openai import OpenAI


class TestOpenAI:
    @pytest.fixture
    def openai_config(self, set_config):
        config = {
            "shelloracle": {"provider": "OpenAI"},
            "provider": {
                "OpenAI": {"api_key": "sk-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx", "model": "gpt-3.5-turbo"}
            },
        }
        set_config(config)

    @pytest.fixture
    def openai_instance(self, openai_config):
        return OpenAI()

    def test_name(self):
        assert OpenAI.name == "OpenAI"

    def test_api_key(self, openai_instance):
        assert openai_instance.api_key == "sk-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"

    def test_model(self, openai_instance):
        assert openai_instance.model == "gpt-3.5-turbo"

    @pytest.mark.asyncio
    async def test_generate(self, mock_asyncopenai, openai_instance):
        result = ""
        async for response in openai_instance.generate(""):
            result += response
        assert result == "head -c 100 /dev/urandom | hexdump -C"
